from __future__ import annotations

from typing import Optional, cast

from coredis._json import json
from coredis.modules.response.types import JsonType
from coredis.response._callbacks import ResponseCallback
from coredis.typing import ResponseType, ValueT


class JsonCallback(ResponseCallback[ResponseType, ResponseType, JsonType]):
    def transform(self, response: ResponseType, **kwargs: Optional[ValueT]) -> JsonType:
        if isinstance(response, (bytes, str)):
            deser = json.loads(response)
        elif isinstance(response, list):
            deser = [
                json.loads(e) if isinstance(e, (bytes, str)) else e for e in response
            ]
        else:
            deser = response
        return cast(JsonType, deser)  # alas we lie.



from __future__ import annotations

from coredis._utils import b
from coredis.typing import Final

SYM_STAR: Final[bytes] = b("*")
SYM_DOLLAR: Final[bytes] = b("$")
SYM_CRLF: Final[bytes] = b("\r\n")
SYM_LF: Final[bytes] = b("\n")
SYM_EMPTY: Final[bytes] = b("")


class RESPDataType:
    """
    Markers used by redis server to signal
    the type of data being sent.

    See:

    - `RESP protocol spec <https://redis.io/docs/reference/protocol-spec/>`__
    - `RESP3 specification <https://github.com/antirez/RESP3/blob/master/spec.md>`__
    """

    NONE: Final[int] = ord(b"_")
    SIMPLE_STRING: Final[int] = ord(b"+")
    BULK_STRING: Final[int] = ord(b"$")
    VERBATIM: Final[int] = ord(b"=")
    BOOLEAN: Final[int] = ord(b"#")
    INT: Final[int] = ord(b":")
    DOUBLE: Final[int] = ord(b",")
    BIGNUMBER: Final[int] = ord(b"(")
    ARRAY: Final[int] = ord(b"*")
    PUSH: Final[int] = ord(b">")
    MAP: Final[int] = ord(b"%")
    SET: Final[int] = ord(b"~")
    ERROR: Final[int] = ord(b"-")
    ATTRIBUTE: Final[int] = ord(b"|")



"""
coredis
-------

coredis is an async redis client with support for redis server,
cluster & sentinel.
"""

from __future__ import annotations

from typing import cast

from coredis.client import KeyDB, KeyDBCluster, Redis, RedisCluster
from coredis.config import Config
from coredis.connection import (
    BaseConnection,
    ClusterConnection,
    Connection,
    UnixDomainSocketConnection,
)
from coredis.pool import (
    BlockingClusterConnectionPool,
    BlockingConnectionPool,
    ClusterConnectionPool,
    ConnectionPool,
)
from coredis.tokens import PureToken

from . import _version

__all__ = [
    "Config",
    "KeyDB",
    "KeyDBCluster",
    "Redis",
    "RedisCluster",
    "BaseConnection",
    "Connection",
    "UnixDomainSocketConnection",
    "ClusterConnection",
    "BlockingConnectionPool",
    "ConnectionPool",
    "BlockingClusterConnectionPool",
    "ClusterConnectionPool",
    "PureToken",
]

__version__ = cast(str, _version.get_versions()["version"])  # type: ignore



from __future__ import annotations

import asyncio
import dataclasses
import time
import weakref
from abc import ABC, abstractmethod
from collections import Counter
from typing import TYPE_CHECKING, Any

from coredis._sidecar import Sidecar
from coredis._utils import b, make_hashable
from coredis.commands import PubSub
from coredis.connection import BaseConnection
from coredis.typing import (
    Dict,
    Generic,
    Hashable,
    List,
    Literal,
    Optional,
    OrderedDict,
    Protocol,
    ResponseType,
    Tuple,
    TypeVar,
    Union,
    ValueT,
    runtime_checkable,
)

try:
    from pympler import asizeof
# Not available in pypy
except AttributeError:
    asizeof = None

if TYPE_CHECKING:
    import coredis.client


@dataclasses.dataclass
class CacheStats:
    """
    Summary of statics to be used by instances of :class:`coredis.cache.AbstractCache`
    The individual counters exposed are not guaranteed to retain fine grained per key
    metrics but the totals (returned by :attr:`coredis.cache.CacheStats.summary`) will be maintained
    aggregated.
    """

    #: summary of hits by key (for all commands)
    hits: Counter[bytes] = dataclasses.field(default_factory=Counter)
    #: summary of misses by key (for all commands)
    misses: Counter[bytes] = dataclasses.field(default_factory=Counter)
    #: number of invalidations including server side and local invalidations
    invalidations: Counter[bytes] = dataclasses.field(default_factory=Counter)
    #: counter of keys which returned dirty results based on confidence testing
    dirty: Counter[bytes] = dataclasses.field(default_factory=Counter)

    def clear(self) -> None:
        self.hits.clear()
        self.misses.clear()
        self.invalidations.clear()
        self.dirty.clear()

    def compact(self) -> None:
        """
        Collapse totals into a single key to avoid unbounded growth of stats

        :meta private:
        """

        for counter in [self.hits, self.misses, self.invalidations, self.dirty]:
            total = sum(counter.values())
            counter.clear()
            counter[b"__coredis__internal__stats__total"] = total

    def hit(self, key: ValueT) -> None:
        self.hits[b(key)] += 1

    def miss(self, key: ValueT) -> None:
        self.misses[b(key)] += 1

    def invalidate(self, key: ValueT) -> None:
        self.invalidations[b(key)] += 1

    def mark_dirty(self, key: ValueT) -> None:
        self.dirty[b(key)] += 1

    @property
    def summary(self) -> Dict[str, int]:
        """
        Aggregated totals of ``hits``, ``misses``, ``dirty_hits``
        and ``invalidations``
        """

        return {
            "hits": sum(self.hits.values()),
            "misses": sum(self.misses.values()),
            "dirty_hits": sum(self.dirty.values()),
            "invalidations": sum(self.invalidations.values()),
        }

    def __repr__(self) -> str:
        summary = self.summary

        return (
            f"CacheStats<hits={summary['hits']}, "
            f"misses={summary['misses']}, "
            f"dirty_hits={summary['dirty_hits']}, "
            f"invalidations={summary['invalidations']}>"
        )


class AbstractCache(ABC):
    """
    Abstract class representing a local cache that can be used by
    :class:`coredis.Redis` or :class:`coredis.RedisCluster`
    """

    @abstractmethod
    async def initialize(
        self,
        client: Union["coredis.client.Redis[Any]", "coredis.client.RedisCluster[Any]"],
    ) -> AbstractCache:
        """
        Associate and initialize this cache with the provided client
        """
        ...

    @property
    @abstractmethod
    def healthy(self) -> bool:
        """
        Whether the cache is healthy and should be taken seriously
        """
        ...

    @abstractmethod
    def get(self, command: bytes, key: bytes, *args: ValueT) -> ResponseType:
        """
        Fetch the cached response for command/key/args combination
        """
        ...

    @abstractmethod
    def put(
        self, command: bytes, key: bytes, *args: ValueT, value: ResponseType
    ) -> None:
        """
        Cache the response for command/key/args combination
        """
        ...

    @abstractmethod
    def invalidate(self, *keys: ValueT) -> None:
        """
        Invalidate any cached entries for the provided keys
        """
        ...

    @abstractmethod
    def reset(self) -> None:
        """
        Reset the cache
        """
        ...

    @abstractmethod
    def shutdown(self) -> None:
        """
        Explicitly shutdown the cache
        """
        ...


@runtime_checkable
class SupportsStats(Protocol):
    """
    Protocol of a cache that provides cache statistics
    """

    @property
    @abstractmethod
    def stats(self) -> CacheStats:
        """
        Returns the current stats for the cache
        """
        ...


@runtime_checkable
class SupportsSampling(Protocol):
    """
    If a cache implements :class:`SupportsSampling`, methods that support
    caching will sample the response from the cache and test it against an uncached
    response from the server based on the confidence returned by :attr:`confidence`.
    The outcome of the validation will be fed back to the cache using :meth:`feedback`
    and in the case that there was no match, the uncached response will be returned.
    """

    @property
    @abstractmethod
    def confidence(self) -> float:
        """
        Confidence in cached values between 0 - 100. Lower values
        will result in the client discarding and / or validating the
        cached responses
        """
        ...

    @abstractmethod
    def feedback(self, command: bytes, key: bytes, *args: ValueT, match: bool) -> None:
        """
        Provide feedback about a key as having either a match or drift from the actual
        server side value
        """
        ...


@runtime_checkable
class SupportsClientTracking(Protocol):
    """
    If a cache implements :class:`SupportsClientTracking`, the :class:`~coredis.Redis`
    and :class:`~coredis.RedisCluster` clients will ensure that the client
    returned by :meth:`get_client_id` is set using the :meth:`~coredis.Redis.client_tracking`
    command on any connection returned by the clients.
    """

    @abstractmethod
    def get_client_id(self, connection: BaseConnection) -> Optional[int]:
        """
        If the cache supports receiving invalidation events from the server
        return the ``client_id`` that the :paramref:`connection` should send
        redirects to.
        """
        ...


ET = TypeVar("ET")


class LRUCache(Generic[ET]):
    def __init__(self, max_items: int = -1, max_bytes: int = -1):
        self.max_items = max_items
        self.max_bytes = max_bytes
        self.__cache: OrderedDict[Hashable, ET] = OrderedDict()

        if self.max_bytes > 0 and asizeof is not None:
            self.max_bytes += asizeof.asizeof(self.__cache)
        elif self.max_bytes > 0:
            raise RuntimeError(
                "max_bytes not supported as dependency pympler not available"
            )

    def get(self, key: Hashable) -> ET:
        if key not in self.__cache:
            raise KeyError(key)
        self.__cache.move_to_end(key)

        return self.__cache[key]

    def insert(self, key: Hashable, value: ET) -> None:
        self.__check_capacity()
        self.__cache[key] = value
        self.__cache.move_to_end(key)

    def setdefault(self, key: Hashable, value: ET) -> ET:
        try:
            self.__check_capacity()

            return self.get(key)
        except KeyError:
            self.insert(key, value)

            return self.get(key)

    def remove(self, key: Hashable) -> None:
        if key in self.__cache:
            self.__cache.pop(key)

    def clear(self) -> None:
        self.__cache.clear()

    def popitem(self) -> bool:
        """
        Recursively remove the oldest entry. If
        the oldest entry is another LRUCache trigger
        the removal of its oldest entry and if that
        turns out to be an empty LRUCache, remove that.
        """
        try:
            oldest = next(iter(self.__cache))
            item = self.__cache[oldest]
        except StopIteration:
            return False

        if isinstance(item, LRUCache):
            if item.popitem():
                return True
        self.__cache.popitem(last=False)

        return True

    def shrink(self) -> None:
        """
        Remove old entries until the size of the cache
        is less than :paramref:`LRUCache.max_bytes` or if
        there is nothing left to remove.
        """

        if self.max_bytes > 0 and asizeof is not None:
            while asizeof.asizeof(self.__cache) > self.max_bytes:
                if not self.popitem():
                    # nothing left to remove

                    return

    def __repr__(self) -> str:
        if asizeof is not None:
            return (
                f"LruCache<max_items={self.max_items}, "
                f"current_items={len(self.__cache)}, "
                f"max_bytes={self.max_bytes}, "
                f"current_size_bytes={asizeof.asizeof(self)}>"
            )
        else:
            return (
                f"LruCache<max_items={self.max_items}, "
                f"current_items={len(self.__cache)}, "
            )

    def __check_capacity(self) -> None:
        if len(self.__cache) == self.max_items:
            self.__cache.popitem(last=False)


class NodeTrackingCache(
    Sidecar,
    AbstractCache,
    SupportsStats,
    SupportsSampling,
    SupportsClientTracking,
):
    """
    An LRU cache that uses server assisted client caching
    to ensure local cache entries are invalidated if any
    operations are performed on the keys by another client.
    """

    def __init__(
        self,
        max_keys: int = 2**12,
        max_size_bytes: int = 64 * 1024 * 1024,
        max_idle_seconds: int = 5,
        confidence: float = 100,
        dynamic_confidence: bool = False,
        cache: Optional[LRUCache[LRUCache[LRUCache[ResponseType]]]] = None,
        stats: Optional[CacheStats] = None,
    ) -> None:
        """
        :param max_keys: maximum keys to cache. A negative value represents
         and unbounded cache.
        :param max_size_bytes: maximum size in bytes for the local cache.
         A negative value represents an unbounded cache.
        :param max_idle_seconds: maximum duration to tolerate no updates
         from the server. When the duration is exceeded the connection
         and cache will be reset.
        :param confidence: 0 - 100. Lower values will result in the client
         discarding and / or validating the cached responses
        :param dynamic_confidence: Whether to adjust the confidence based on
         sampled validations. Tainted values drop the confidence by 0.1% and
         confirmations of correct cached values will increase the confidence by 0.01%
         upto 100.
        """
        super().__init__({b"invalidate"}, max(1, max_idle_seconds - 1))
        self.__protocol_version: Optional[Literal[2, 3]] = None
        self.__invalidation_task: Optional[asyncio.Task[None]] = None
        self.__compact_task: Optional[asyncio.Task[None]] = None
        self.__max_idle_seconds = max_idle_seconds
        self.__confidence = self.__original_confidence = confidence
        self.__dynamic_confidence = dynamic_confidence
        self.__stats = stats or CacheStats()
        self.__cache: LRUCache[LRUCache[LRUCache[ResponseType]]] = cache or LRUCache(
            max_keys, max_size_bytes
        )

    @property
    def healthy(self) -> bool:
        return bool(
            self.connection
            and self.connection.is_connected
            and time.monotonic() - self.last_checkin < self.__max_idle_seconds
        )

    @property
    def confidence(self) -> float:
        return self.__confidence

    @property
    def stats(self) -> CacheStats:
        return self.__stats

    def get(self, command: bytes, key: bytes, *args: ValueT) -> ResponseType:
        try:
            cached = self.__cache.get(b(key)).get(command).get(make_hashable(*args))
            self.__stats.hit(key)

            return cached
        except KeyError:
            self.__stats.miss(key)
            raise

    def put(
        self, command: bytes, key: bytes, *args: ValueT, value: ResponseType
    ) -> None:
        self.__cache.setdefault(b(key), LRUCache()).setdefault(
            command, LRUCache()
        ).insert(make_hashable(*args), value)

    def invalidate(self, *keys: ValueT) -> None:
        for key in keys:
            self.__stats.invalidate(key)
            self.__cache.remove(b(key))

    def feedback(self, command: bytes, key: bytes, *args: ValueT, match: bool) -> None:
        if not match:
            self.__stats.mark_dirty(key)
            self.invalidate(key)

        if self.__dynamic_confidence:
            self.__confidence = min(
                100.0,
                max(0.0, self.__confidence * (1.0001 if match else 0.999)),
            )

    def reset(self) -> None:
        self.__cache.clear()
        self.__stats.compact()
        self.__confidence = self.__original_confidence

    def process_message(self, message: ResponseType) -> Tuple[ResponseType, ...]:
        assert isinstance(message, list)

        if self.__protocol_version == 2:
            assert isinstance(message[0], bytes)

            if b(message[0]) in PubSub.SUBUNSUB_MESSAGE_TYPES:
                return ()
            elif message[2] is not None:
                assert isinstance(message[2], list)

                return tuple(k for k in message[2])
        elif message[1] is not None:
            assert isinstance(message[1], list)

            return tuple(k for k in message[1])

        return ()  # noqa

    async def initialize(
        self,
        client: Union["coredis.client.Redis[Any]", "coredis.client.RedisCluster[Any]"],
    ) -> NodeTrackingCache:
        self.__protocol_version = client.protocol_version
        await super().start(client)

        if not self.__invalidation_task or self.__invalidation_task.done():
            self.__invalidation_task = asyncio.create_task(self.__invalidate())

        if not self.__compact_task or self.__compact_task.done():
            self.__compact_task = asyncio.create_task(self.__compact())

        return self

    async def on_reconnect(self, connection: BaseConnection) -> None:
        self.__cache.clear()
        await super().on_reconnect(connection)

        if self.__protocol_version == 2 and self.connection:
            await self.connection.send_command(b"SUBSCRIBE", b"__redis__:invalidate")

    def shutdown(self) -> None:
        try:
            asyncio.get_running_loop()

            if self.__invalidation_task:
                self.__invalidation_task.cancel()

            if self.__compact_task:
                self.__compact_task.cancel()
            super().stop()
        except RuntimeError:
            pass

    def get_client_id(self, client: BaseConnection) -> Optional[int]:
        if self.connection and self.connection.is_connected:
            return self.client_id

        return None

    async def __compact(self) -> None:
        while True:
            try:
                self.__cache.shrink()
                self.__stats.compact()
                await asyncio.sleep(max(1, self.__max_idle_seconds - 1))
            except asyncio.CancelledError:
                break

    async def __invalidate(self) -> None:
        while True:
            try:
                key = b(await self.messages.get())
                self.invalidate(key)
                self.messages.task_done()
            except asyncio.CancelledError:
                break
            except RuntimeError:  # noqa
                break


class ClusterTrackingCache(
    AbstractCache, SupportsStats, SupportsSampling, SupportsClientTracking
):
    """
    An LRU cache for redis cluster that uses server assisted client caching
    to ensure local cache entries are invalidated if any operations are performed
    on the keys by another client.

    The cache maintains an additional connection per node (including replicas)
    in the cluster to listen to invalidation events
    """

    def __init__(
        self,
        max_keys: int = 2**12,
        max_size_bytes: int = 64 * 1024 * 1024,
        max_idle_seconds: int = 5,
        confidence: float = 100,
        dynamic_confidence: bool = False,
        cache: Optional[LRUCache[LRUCache[LRUCache[ResponseType]]]] = None,
        stats: Optional[CacheStats] = None,
    ) -> None:
        """
        :param max_keys: maximum keys to cache. A negative value represents
         and unbounded cache.
        :param max_size_bytes: maximum size in bytes for the local cache.
         A negative value represents an unbounded cache.
        :param max_idle_seconds: maximum duration to tolerate no updates
         from the server. When the duration is exceeded the connection
         and cache will be reset.
        :param confidence: 0 - 100. Lower values will result in the client
         discarding and / or validating the cached responses
        :param dynamic_confidence: Whether to adjust the confidence based on
         sampled validations. Tainted values drop the confidence by 0.1% and
         confirmations of correct cached values will increase the confidence by 0.01%
         upto 100.
        """
        self.node_caches: Dict[str, NodeTrackingCache] = {}
        self.__protocol_version: Optional[Literal[2, 3]] = None
        self.__cache: LRUCache[LRUCache[LRUCache[ResponseType]]] = cache or LRUCache(
            max_keys, max_size_bytes
        )
        self.__nodes: List["coredis.client.Redis[Any]"] = []
        self.__max_idle_seconds = max_idle_seconds
        self.__confidence = self.__original_confidence = confidence
        self.__dynamic_confidence = dynamic_confidence
        self.__stats = stats or CacheStats()
        self.__client: Optional[
            weakref.ReferenceType["coredis.client.RedisCluster[Any]"]
        ] = None

    async def initialize(
        self,
        client: Union["coredis.client.Redis[Any]", "coredis.client.RedisCluster[Any]"],
    ) -> ClusterTrackingCache:
        import coredis.client

        assert isinstance(client, coredis.client.RedisCluster)

        self.__client = weakref.ref(client)
        self.__cache.clear()

        for sidecar in self.node_caches.values():
            sidecar.shutdown()
        self.node_caches.clear()
        self.__nodes = list(client.all_nodes)

        for node in self.__nodes:
            node_cache = NodeTrackingCache(
                max_idle_seconds=self.__max_idle_seconds,
                confidence=self.__confidence,
                dynamic_confidence=self.__dynamic_confidence,
                cache=self.__cache,
                stats=self.__stats,
            )
            await node_cache.initialize(node)
            assert node_cache.connection
            self.node_caches[node_cache.connection.location] = node_cache

        return self

    @property
    def client(self) -> Optional["coredis.client.RedisCluster[Any]"]:
        if self.__client:
            return self.__client()

        return None  # noqa

    @property
    def healthy(self) -> bool:
        return bool(
            self.client
            and self.client.connection_pool.initialized
            and self.node_caches
            and all(cache.healthy for cache in self.node_caches.values())
        )

    @property
    def confidence(self) -> float:
        return self.__confidence

    @property
    def stats(self) -> CacheStats:
        return self.__stats

    def get_client_id(self, connection: BaseConnection) -> Optional[int]:
        try:
            return self.node_caches[connection.location].get_client_id(connection)
        except KeyError:
            return None

    def get(self, command: bytes, key: bytes, *args: ValueT) -> ResponseType:
        try:
            cached = self.__cache.get(b(key)).get(command).get(make_hashable(*args))
            self.__stats.hit(key)

            return cached
        except KeyError:
            self.__stats.miss(key)
            raise

    def put(
        self, command: bytes, key: bytes, *args: ValueT, value: ResponseType
    ) -> None:
        self.__cache.setdefault(b(key), LRUCache()).setdefault(
            command, LRUCache()
        ).insert(make_hashable(*args), value)

    def invalidate(self, *keys: ValueT) -> None:
        for key in keys:
            self.__stats.invalidate(key)
            self.__cache.remove(b(key))

    def feedback(self, command: bytes, key: bytes, *args: ValueT, match: bool) -> None:
        if not match:
            self.__stats.mark_dirty(key)
            self.invalidate(key)

        if self.__dynamic_confidence:
            self.__confidence = min(
                100.0,
                max(0.0, self.__confidence * (1.0001 if match else 0.999)),
            )

    def reset(self) -> None:
        self.__cache.clear()
        self.__stats.compact()
        self.__confidence = self.__original_confidence

    def shutdown(self) -> None:
        if self.node_caches:
            for sidecar in self.node_caches.values():
                sidecar.shutdown()
            self.node_caches.clear()
            self.__nodes.clear()

    def __del__(self) -> None:
        self.shutdown()


class TrackingCache(
    AbstractCache, SupportsStats, SupportsSampling, SupportsClientTracking
):
    """
    An LRU cache that uses server assisted client caching to ensure local cache entries
    are invalidated if any operations are performed on the keys by another client.

    This class proxies to either :class:`~coredis.cache.NodeTrackingCache`
    or :class:`~coredis.cache.ClusterTrackingCache` depending on which type of client
    it is passed into.
    """

    def __init__(
        self,
        max_keys: int = 2**12,
        max_size_bytes: int = 64 * 1024 * 1024,
        max_idle_seconds: int = 5,
        confidence: float = 100.0,
        dynamic_confidence: bool = False,
        cache: Optional[LRUCache[LRUCache[LRUCache[ResponseType]]]] = None,
        stats: Optional[CacheStats] = None,
    ) -> None:
        """
        :param max_keys: maximum keys to cache. A negative value represents
         and unbounded cache.
        :param max_size_bytes: maximum size in bytes for the local cache.
         A negative value represents an unbounded cache.
        :param max_idle_seconds: maximum duration to tolerate no updates
         from the server. When the duration is exceeded the connection
         and cache will be reset.
        :param confidence: 0 - 100. Lower values will result in the client
         discarding and / or validating the cached responses
        :param dynamic_confidence: Whether to adjust the confidence based on
         sampled validations. Tainted values drop the confidence by 0.1% and
         confirmations of correct cached values will increase the confidence by 0.01%
         upto 100.
        """
        self.instance: Optional[Union[ClusterTrackingCache, NodeTrackingCache]] = None
        self.__max_keys = max_keys
        self.__max_size_bytes = max_size_bytes
        self.__max_idle_seconds = max_idle_seconds
        self.__confidence = confidence
        self.__dynamic_confidence = dynamic_confidence
        self.__cache: LRUCache[LRUCache[LRUCache[ResponseType]]] = cache or LRUCache(
            max_keys, max_size_bytes
        )
        self.__client: Optional[
            weakref.ReferenceType[
                Union["coredis.client.Redis[Any]", "coredis.client.RedisCluster[Any]"],
            ]
        ] = None
        self.__stats = stats or CacheStats()

    async def initialize(
        self,
        client: Union["coredis.client.Redis[Any]", "coredis.client.RedisCluster[Any]"],
    ) -> TrackingCache:
        import coredis.client

        if self.__client and self.__client() != client:
            copy = self.share()

            return await copy.initialize(client)

        self.__client = weakref.ref(client)

        if not self.instance:
            if isinstance(client, coredis.client.RedisCluster):
                self.instance = ClusterTrackingCache(
                    self.__max_keys,
                    self.__max_size_bytes,
                    self.__max_idle_seconds,
                    confidence=self.__confidence,
                    dynamic_confidence=self.__dynamic_confidence,
                    cache=self.__cache,
                    stats=self.__stats,
                )
            else:
                self.instance = NodeTrackingCache(
                    self.__max_keys,
                    self.__max_size_bytes,
                    self.__max_idle_seconds,
                    confidence=self.__confidence,
                    dynamic_confidence=self.__dynamic_confidence,
                    cache=self.__cache,
                    stats=self.__stats,
                )
        await self.instance.initialize(client)

        return self

    @property
    def healthy(self) -> bool:
        return bool(self.instance and self.instance.healthy)

    @property
    def confidence(self) -> float:
        if not self.instance:
            return self.__confidence

        return self.instance.confidence

    @property
    def stats(self) -> CacheStats:
        return self.__stats

    def get_client_id(self, connection: BaseConnection) -> Optional[int]:
        if self.instance:
            return self.instance.get_client_id(connection)

        return None

    def get(self, command: bytes, key: bytes, *args: ValueT) -> ResponseType:
        assert self.instance

        return self.instance.get(command, key, *args)

    def put(
        self, command: bytes, key: bytes, *args: ValueT, value: ResponseType
    ) -> None:
        if self.instance:
            self.instance.put(command, key, *args, value=value)

    def invalidate(self, *keys: ValueT) -> None:
        if self.instance:
            self.instance.invalidate(*keys)

    def feedback(self, command: bytes, key: bytes, *args: ValueT, match: bool) -> None:
        if self.instance:
            self.instance.feedback(command, key, *args, match=match)

    def reset(self) -> None:
        if self.instance:
            self.instance.reset()

    def shutdown(self) -> None:
        if self.instance:
            self.instance.shutdown()
        self.__client = None

    def share(self) -> TrackingCache:
        """
        Create a copy of this cache that can be used to share
        memory with another client.

        In the example below ``c1`` and ``c2`` have their own
        instances of :class:`~coredis.cache.TrackingCache` but
        share the same in-memory local cached responses::

            c1 = await coredis.Redis(cache=TrackingCache())
            c2 = await coredis.Redis(cache=c1.cache.share())
        """
        copy = self.__class__(
            self.__max_keys,
            self.__max_size_bytes,
            self.__max_idle_seconds,
            self.__confidence,
            self.__dynamic_confidence,
            self.__cache,
            self.__stats,
        )

        return copy

    def __del__(self) -> None:
        self.shutdown()



from __future__ import annotations

from coredis.constants import SYM_CRLF, SYM_DOLLAR, SYM_EMPTY, SYM_STAR
from coredis.typing import List, Tuple, ValueT


class Packer:
    def __init__(self, encoding: str):
        self.encoding = encoding

    def encode(self, value: ValueT) -> bytes:
        """Returns a bytestring representation of the value"""
        if isinstance(value, str):
            return value.encode(self.encoding)
        elif isinstance(value, int):
            return b"%d" % value
        elif isinstance(value, float):
            return b"%.15g" % value
        return value

    def pack_command(self, command: bytes, *args: ValueT) -> List[bytes]:
        "Pack a series of arguments into the Redis protocol"
        output: List[bytes] = []
        # the client might have included 1 or more literal arguments in
        # the command name, e.g., 'CONFIG GET'. The Redis server expects these
        # arguments to be sent separately, so split the first argument
        # manually. All of these arguements get wrapped in the Token class
        # to prevent them from being encoded.
        cleaned_args = args
        if b" " in command:
            cleaned_args = tuple(s for s in command.split()) + cleaned_args
        else:
            cleaned_args = (command,) + cleaned_args

        buff = SYM_EMPTY.join((SYM_STAR, b"%d" % len(cleaned_args), SYM_CRLF))

        for arg in cleaned_args:
            if not isinstance(arg, bytes):
                arg = self.encode(arg)
            # to avoid large string mallocs, chunk the command into the
            # output list if we're sending large values

            if len(buff) > 6000 or len(arg) > 6000:
                buff = SYM_EMPTY.join((buff, SYM_DOLLAR, b"%d" % len(arg), SYM_CRLF))
                output.append(buff)
                output.append(arg)
                buff = SYM_CRLF
            else:
                buff = SYM_EMPTY.join(
                    (buff, SYM_DOLLAR, b"%d" % len(arg), SYM_CRLF, arg, SYM_CRLF)
                )
        output.append(buff)
        return output

    def pack_commands(self, commands: List[Tuple[ValueT, ...]]) -> List[bytes]:
        output: List[bytes] = []
        pieces: List[bytes] = []
        buffer_length = 0

        for cmd in commands:
            for chunk in self.pack_command(self.encode(cmd[0]), *cmd[1:]):
                pieces.append(chunk)
                buffer_length += len(chunk)

            if buffer_length > 6000:
                output.append(SYM_EMPTY.join(pieces))
                buffer_length = 0
                pieces = []

        if pieces:
            output.append(SYM_EMPTY.join(pieces))

        return output



from __future__ import annotations

import functools
import inspect
import itertools
import weakref
from typing import Any, ClassVar, cast

from deprecated.sphinx import versionadded

from coredis._utils import EncodingInsensitiveDict, nativestr
from coredis.exceptions import FunctionError
from coredis.typing import (
    TYPE_CHECKING,
    AnyStr,
    Awaitable,
    Callable,
    Dict,
    Generator,
    Generic,
    KeyT,
    List,
    Optional,
    P,
    Parameters,
    R,
    ResponseType,
    StringT,
    Tuple,
    TypeVar,
    ValueT,
    add_runtime_checks,
    safe_beartype,
)

if TYPE_CHECKING:
    import coredis.client

LibraryT = TypeVar("LibraryT", bound="Library[Any]")
LibraryStringT = TypeVar("LibraryStringT", bound="Library[str]")
LibraryBytesT = TypeVar("LibraryBytesT", bound="Library[bytes]")


class Library(Generic[AnyStr]):
    #: Class variable equivalent of the :paramref:`Library.name` argument.
    NAME: ClassVar[Optional[StringT]] = None
    #: Class variable equivalent of the :paramref:`Library.code` argument.
    CODE: ClassVar[Optional[StringT]] = None

    def __init__(
        self,
        client: coredis.client.Client[AnyStr],
        name: Optional[StringT] = None,
        code: Optional[StringT] = None,
        replace: bool = False,
    ) -> None:
        """
        Abstraction over a library of redis functions

        Example::

            library_code = \"\"\"
            #!lua name=coredis
            redis.register_function('myfunc', function(k, a) return a[1] end)
            \"\"\"
            lib = await Library(client, "mylib", library_code)
            assert "1" == await lib["myfunc"]([], [1])

        When used as a base class the class variables :data:`NAME` and :data:`CODE`
        can be set on the sub class to avoid having to implement a constructor. Constructor
        parameters will take precedence over the class variables.

        :param client: The coredis client instance to use when calling the functions
         exposed by the library.
        :param name: The name of the library (should match the name in the Shebang
         in the library source).
        :param code: The lua code representing the library
        :param replace: Whether to replace the library when intializing. If ``False``
         an exception will be raised if the library was already loaded in the target
         redis instance.
        """
        self._client: weakref.ReferenceType[
            coredis.client.Client[AnyStr]
        ] = weakref.ref(client)
        self.name = nativestr(name or self.NAME)
        self.code = (code or self.CODE or "").lstrip()
        self._functions: EncodingInsensitiveDict = EncodingInsensitiveDict()
        self.replace = replace
        if self.replace and not self.code:
            raise RuntimeError(
                "library code must be provided when the ``replace`` option is used"
            )

    @property
    def client(self) -> coredis.client.Client[AnyStr]:
        c = self._client()
        assert c
        return c

    @property
    def functions(self) -> Dict[str, Function[AnyStr]]:
        """
        mapping of function names to :class:`~coredis.commands.function.Function`
        instances that can be directly called.
        """
        return self._functions

    async def update(self, new_code: StringT) -> bool:
        """
        Update the code of a library with :paramref:`new_code`
        """
        self.code = new_code
        if await self.initialize(replace=True):
            return True
        return False

    async def initialize(self: LibraryT, replace: bool = False) -> LibraryT:
        self._functions.clear()
        library = (await self.client.function_list(self.name)).get(self.name)
        if (not library and self.code) or (replace or self.replace):
            await self.client.function_load(self.code, replace=replace or self.replace)
            library = (await self.client.function_list(self.name)).get(self.name)

        if not library:
            raise FunctionError(f"No library found for {self.name}")

        for name, details in library["functions"].items():
            self._functions[name] = Function[AnyStr](
                self.client,
                self.name,
                name,
                bool({b"no-writes", "no-writes"} & details["flags"]),
            )
        return self

    def __await__(self: LibraryT) -> Generator[Any, None, LibraryT]:
        return self.initialize().__await__()

    def __getitem__(self, function: str) -> Optional[Function[AnyStr]]:
        return cast(Optional[Function[AnyStr]], self._functions.get(function))

    @classmethod
    @versionadded(version="3.5.0")
    def wraps(
        cls,
        function_name: str,
        key_spec: Optional[List[KeyT]] = None,
        param_is_key: Callable[[inspect.Parameter], bool] = lambda p: (
            p.annotation in {"KeyT", KeyT}
        ),
        runtime_checks: bool = False,
        readonly: Optional[bool] = None,
    ) -> Callable[[Callable[P, Awaitable[R]]], Callable[P, Awaitable[R]]]:
        """
        Decorator for wrapping methods of subclasses of :class:`Library`
        as entry points to the functions contained in the library. This allows
        exposing a strict signature instead of that which :meth:`Function.__call__`
        provides. The callable being decorated should **not** have an implementation as
        it will never be called.

        The main objective of the decorator is to allow you to represent a lua library of
        functions as a python class having strict (and type safe) methods as entry points.
        Internally the decorator separates ``keys`` from ``args`` before calling
        :meth:`coredis.Redis.fcall`.

        Mapping the decorated method's arguments to key providers is done either by
        using :paramref:`key_spec` or :paramref:`param_is_key`. All other parameters of the
        decorated method are assumed to be ``args`` consumed by the lua function.


        The following example demonstrates most of the functionality provided by the
        decorator::

            import coredis
            from coredis.commands import Library
            from coredis.typing import KeyT, ValueT
            from typing import List

            class MyAwesomeLibrary(Library):
                NAME = "mylib"
                CODE = \"\"\"
                #!lua name=mylib

                redis.register_function('echo', function(k, a)
                    return a[1]
                end)
                redis.register_function('ping', function()
                    return "PONG"
                end)
                redis.register_function('get', function(k, a)
                    return redis.call("GET", k[1])
                end)
                redis.register_function('hmget', function(k, a)
                    local values = {}
                    local fields = {}
                    local response = {}
                    local i = 1
                    local j = 1

                    while a[i] do
                        fields[j] = a[i]
                        i = i + 2
                        j = j + 1
                    end

                    for idx, key in ipairs(k) do
                        values = redis.call("HMGET", key, unpack(fields))
                        for idx, value in ipairs(values) do
                            if not response[idx] and value then
                                response[idx] = value
                            end
                        end
                    end
                    for idx, value in ipairs(fields) do
                        if not response[idx] then
                            response[idx] = a[idx*2]
                        end
                    end
                    return response
                end)
                \"\"\"

                @Library.wraps("echo")
                async def echo(self, value: ValueT) -> ValueT: ...

                @Library.wraps("ping")
                async def ping(self) -> str: ...

                @Library.wraps("get")
                async def get(self, key: KeyT) -> ValueT: ...

                @Library.wraps("hmmget")
                async def hmmget(self, *keys: KeyT, **fields_with_values: ValueT): ...
                \"\"\"
                Return values of ``fields_with_values`` on a first come first serve
                basis from the hashes at ``keys``. Since ``fields_with_values`` is a mapping
                the keys are mapped to hash fields and the values are used
                as defaults if they are not found in any of the hashes at ``keys``
                \"\"\"

            client = coredis.Redis()
            lib = await MyAwesomeLibrary(client, replace=True)
            await client.set("hello", "world")
            # True
            await lib.echo("hello world")
            # b"hello world"
            await lib.ping()
            # b"pong"
            await lib.get("hello")
            # b"hello"
            await client.hset("k1", {"c": 3, "d": 4})
            await client.hset("k2", {"a": 1, "b": 2})
            await lib.hmmget("k1", "k2", a=-1, b=-2, c=-3, d=-4, e=-5)
            # [b"1", b"2", b"3", b"4", b"-5"]

        :param key_spec: list of parameters of the decorated method that will
         be passed as the :paramref:`keys` argument to :meth:`__call__`. If provided
         this parameter takes precedence over using :paramref:`param_is_key` to
         determine if a parameter is a key provider.
        :param param_is_key: a callable that accepts a single argument of type
         :class:`inspect.Parameter` and returns ``True`` if the parameter points to a key
         that should be appended to the :paramref:`__call__.keys` argument of
         :meth:`__call__`. The default implementation marks a parameter as a key
         provider if it is of type :data:`coredis.typing.KeyT` and is only used
         if :paramref:`key_spec` is ``None``.
        :param runtime_checks: Whether to enable runtime type checking of input arguments
         and return values. (requires :pypi:`beartype`). If :data:`False` the function will
         still get runtime type checking if the environment configuration ``COREDIS_RUNTIME_CHECKS``
         is set - for details see :ref:`handbook/typing:runtime type checking`.
        :param readonly: If ``True`` forces this function to use :meth:`coredis.Redis.fcall_ro`

        :return: A function that has a signature mirroring the decorated function.
        """

        def wrapper(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]:
            sig = inspect.signature(func)
            first_arg: str = list(sig.parameters.keys())[0]
            runtime_check_wrapper = (
                add_runtime_checks if not runtime_checks else safe_beartype
            )
            key_params = (
                key_spec
                if key_spec
                else [n for n, p in sig.parameters.items() if param_is_key(p)]
            )
            arg_fetch: Dict[str, Callable[..., Parameters[Any]]] = {
                n: (lambda v: [v])
                if p.kind
                in {
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.KEYWORD_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                }
                else (lambda v: list(itertools.chain.from_iterable(v.items())))
                if p.kind == inspect.Parameter.VAR_KEYWORD
                else lambda v: list(v)
                for n, p in sig.parameters.items()
            }

            def split_args(
                *a: P.args, **k: P.kwargs
            ) -> Tuple[Library[AnyStr], Parameters[KeyT], Parameters[ValueT]]:
                bound_arguments = sig.bind(*a, **k)
                bound_arguments.apply_defaults()
                arguments: Dict[str, Any] = bound_arguments.arguments
                instance: Library[AnyStr] = arguments.pop(first_arg)
                if not isinstance(
                    instance, Library
                ):  # pyright: reportUnnecessaryIsInstance=false
                    raise RuntimeError(
                        f"{instance.__class__.__name__} is not a subclass of"
                        " coredis.commands.function.Library therefore it's methods cannot be bound "
                        " to a redis library using ``Library.wrap``."
                        " Please refer to the documentation at https://coredis.readthedocs.org/"
                        " for instructions on how to bind a class to a redis library."
                    )
                keys: List[KeyT] = []
                args: List[ValueT] = []
                for name in sig.parameters:
                    if name == first_arg:
                        continue
                    values = arg_fetch[name](arguments[name])
                    if name in key_params:
                        keys.extend(values)
                    else:
                        args.extend(values)
                return instance, keys, args

            @runtime_check_wrapper
            @functools.wraps(func)
            async def _inner(*args: P.args, **kwargs: P.kwargs) -> R:
                instance, keys, arguments = split_args(*args, **kwargs)
                func = instance.functions[function_name]
                if not func:
                    raise AttributeError(
                        f"Library {instance.name} has no registered function {function_name}"
                    )
                # TODO: atleast lie with a cast.
                #  mypy doesn't like the cast. pyright is ok with it
                return await func(keys, arguments, readonly=readonly)  # type: ignore

            return _inner

        return wrapper


class Function(Generic[AnyStr]):
    def __init__(
        self,
        client: coredis.client.Client[AnyStr],
        library_name: StringT,
        name: StringT,
        readonly: bool = False,
    ):
        """
        Wrapper to call a redis function that has already been loaded

        :param library_name: Name of the library under which the function is registered
        :param name: Name of the function this instance represents
        :param readonly: If ``True`` the function will be called with
         :meth:`coredis.Redis.fcall_ro` instead of :meth:`coredis.Redis.fcall`

        Example::

            func = await Function(client, "mylib", "myfunc")
            response = await func(keys=["a"], args=[1])
        """
        self._client: weakref.ReferenceType[
            coredis.client.Client[AnyStr]
        ] = weakref.ref(client)
        self.library: Library[AnyStr] = Library[AnyStr](client, library_name)
        self.name = name
        self.readonly = readonly

    @property
    def client(self) -> coredis.client.Client[AnyStr]:
        c = self._client()
        assert c
        return c

    async def initialize(self) -> Function[AnyStr]:
        await self.library
        return self

    def __await__(self) -> Generator[Any, None, Function[AnyStr]]:
        return self.initialize().__await__()

    async def __call__(
        self,
        keys: Optional[Parameters[KeyT]] = None,
        args: Optional[Parameters[ValueT]] = None,
        *,
        client: Optional[coredis.client.Client[AnyStr]] = None,
        readonly: Optional[bool] = None,
    ) -> ResponseType:
        """
        Wrapper to call :meth:`~coredis.Redis.fcall` with the
        function named :paramref:`Function.name` registered under
        the library at :paramref:`Function.library`

        :param keys: The keys this function will reference
        :param args: The arguments expected by the function
        :param readonly: If ``True`` forces the function to use :meth:`coredis.Redis.fcall_ro`
        """
        if client is None:
            client = self.client
        if readonly is None:
            readonly = self.readonly

        if readonly:
            return await client.fcall_ro(self.name, keys or [], args or [])
        else:
            return await client.fcall(self.name, keys or [], args or [])



from __future__ import annotations

import asyncio
from io import BytesIO
from typing import Hashable, Type, cast

from coredis._protocols import ConnectionP
from coredis._utils import b
from coredis.constants import SYM_CRLF, RESPDataType
from coredis.exceptions import (
    AskError,
    AuthenticationFailureError,
    AuthenticationRequiredError,
    AuthorizationError,
    BusyLoadingError,
    ClusterCrossSlotError,
    ClusterDownError,
    ConnectionError,
    ExecAbortError,
    InvalidResponse,
    MovedError,
    NoKeyError,
    NoScriptError,
    NotBusyError,
    ProtocolError,
    ReadOnlyError,
    RedisError,
    ResponseError,
    StreamConsumerGroupError,
    StreamDuplicateConsumerGroupError,
    TryAgainError,
    UnblockedError,
    UnknownCommandError,
    WrongTypeError,
)
from coredis.typing import (
    Dict,
    Final,
    FrozenSet,
    List,
    MutableSet,
    NamedTuple,
    Optional,
    ResponsePrimitive,
    ResponseType,
    Set,
    Tuple,
    Union,
)


class NotEnoughData:
    pass


NOT_ENOUGH_DATA: Final[NotEnoughData] = NotEnoughData()


class RESPNode:
    __slots__ = ("depth", "key", "node_type")
    depth: int
    key: Union[
        ResponsePrimitive, Tuple[ResponsePrimitive, ...], FrozenSet[ResponsePrimitive]
    ]
    node_type: int

    def __init__(
        self,
        depth: int,
        node_type: int,
        key: Union[
            ResponsePrimitive,
            Tuple[ResponsePrimitive, ...],
            FrozenSet[ResponsePrimitive],
        ],
    ):
        self.depth = depth
        self.node_type = node_type
        self.key = key

    def append(self, item: ResponseType) -> None:
        raise NotImplementedError()

    def ensure_hashable(self, item: ResponseType) -> Hashable:
        if isinstance(item, (int, float, bool, str, bytes)):
            return item
        elif isinstance(item, set):
            return frozenset(self.ensure_hashable(cast(ResponseType, i)) for i in item)
        elif isinstance(item, list):
            return tuple(self.ensure_hashable(i) for i in item)
        elif isinstance(item, dict):
            return tuple(
                (cast(ResponsePrimitive, k), self.ensure_hashable(v))
                for k, v in item.items()
            )
        return item  # noqa


class ListNode(RESPNode):
    __slots__ = ("container",)

    def __init__(self, depth: int, node_type: int) -> None:
        self.container: List[ResponseType] = []
        super().__init__(depth, node_type, None)

    def append(self, item: ResponseType) -> None:
        self.depth -= 1
        self.container.append(item)


class DictNode(RESPNode):
    __slots__ = ("container",)

    def __init__(self, depth: int) -> None:
        self.container: Dict[
            Union[
                ResponsePrimitive,
                Tuple[ResponsePrimitive, ...],
                FrozenSet[ResponsePrimitive],
            ],
            ResponseType,
        ] = {}
        super().__init__(depth * 2, RESPDataType.MAP, None)

    def append(self, item: ResponseType) -> None:
        self.depth -= 1
        if not self.key:
            self.key = cast(
                Union[
                    ResponsePrimitive,
                    Tuple[ResponsePrimitive, ...],
                    FrozenSet[ResponsePrimitive],
                ],
                self.ensure_hashable(item),
            )
        else:
            self.container[self.key] = item
            self.key = None


class SetNode(RESPNode):
    __slots__ = ("container",)

    def __init__(self, depth: int) -> None:
        self.container: MutableSet[
            Union[
                ResponsePrimitive,
                Tuple[ResponsePrimitive, ...],
                FrozenSet[ResponsePrimitive],
            ]
        ] = set()
        super().__init__(depth, RESPDataType.SET, None)

    def append(
        self,
        item: ResponseType,
    ) -> None:
        self.depth -= 1
        self.container.add(
            cast(
                Union[
                    ResponsePrimitive,
                    Tuple[ResponsePrimitive, ...],
                    FrozenSet[ResponsePrimitive],
                ],
                self.ensure_hashable(item),
            )
        )


class UnpackedResponse(NamedTuple):
    response_type: int
    response: ResponseType


class Parser:
    """
    Interface between a connection and Unpacker
    """

    EXCEPTION_CLASSES: Dict[
        str, Union[Type[RedisError], Dict[str, Type[RedisError]]]
    ] = {
        "ASK": AskError,
        "BUSYGROUP": StreamDuplicateConsumerGroupError,
        "CLUSTERDOWN": ClusterDownError,
        "CROSSSLOT": ClusterCrossSlotError,
        "ERR": {
            "max number of clients reached": ConnectionError,
            "unknown command": UnknownCommandError,
            "unknown subcommand": UnknownCommandError,
            "sub command not supported": UnknownCommandError,
        },
        "EXECABORT": ExecAbortError,
        "LOADING": BusyLoadingError,
        "NOSCRIPT": NoScriptError,
        "MOVED": MovedError,
        "NOAUTH": AuthenticationRequiredError,
        "NOGROUP": StreamConsumerGroupError,
        "NOKEY": NoKeyError,
        "NOPERM": AuthorizationError,
        "NOPROTO": ProtocolError,
        "NOTBUSY": NotBusyError,
        "READONLY": ReadOnlyError,
        "TRYAGAIN": TryAgainError,
        "UNBLOCKED": UnblockedError,
        "WRONGPASS": AuthenticationFailureError,
        "WRONGTYPE": WrongTypeError,
    }

    def __init__(self) -> None:
        self.push_messages: Optional[asyncio.Queue[ResponseType]] = None
        self.localbuffer: BytesIO = BytesIO(b"")
        self.bytes_read: int = 0
        self.bytes_written: int = 0
        self.nodes: List[Union[ListNode, SetNode, DictNode]] = []

    def feed(self, data: bytes) -> None:
        self.localbuffer.seek(self.bytes_written)
        self.bytes_written += self.localbuffer.write(data)
        self.localbuffer.seek(self.bytes_read)

    def on_connect(self, connection: ConnectionP) -> None:
        """Called when the stream connects"""
        self.push_messages = connection.push_messages

    def on_disconnect(self) -> None:
        """Called when the stream disconnects"""
        if not self.localbuffer.closed:
            self.localbuffer.seek(0)
            self.localbuffer.truncate()
            self.bytes_read = self.bytes_written = 0
        self.nodes.clear()

    def can_read(self) -> bool:
        return (self.bytes_written - self.bytes_read) > 0

    def try_decode(self, data: bytes, encoding: str) -> Union[bytes, str]:
        try:
            return data.decode(encoding)
        except ValueError:
            return data

    def get_response(
        self,
        decode: bool,
        encoding: Optional[str] = None,
        push_message_types: Optional[Set[bytes]] = None,
    ) -> Union[NotEnoughData, ResponseType]:
        """

        :param decode: Whether to decode simple or bulk strings
        :param push_message_types: the push message types to return if they
         arrive. If a message arrives that does not match the filter, it will
         be put on the :data:`~coredis.connection.BaseConnection.push_messages`
         queue
        :return: The next available parsed response read from the connection.
         If there is not enough data on the wire a ``NotEnoughData`` instance
         will be returned.
        """
        while True:
            response = self.parse(decode, encoding)
            if isinstance(response, NotEnoughData):
                return response
            else:
                if response and response.response_type == RESPDataType.PUSH:
                    assert isinstance(response.response, list)
                    assert self.push_messages
                    if (
                        not push_message_types
                        or b(response.response[0]) not in push_message_types
                    ):
                        self.push_messages.put_nowait(response.response)
                        continue
                    else:
                        break
                else:
                    break
        return response.response if response else None

    def parse(
        self,
        decode_bytes: bool,
        encoding: Optional[str],
    ) -> Union[Optional[UnpackedResponse], NotEnoughData]:
        parsed: Optional[UnpackedResponse] = None

        while True:
            data = self.localbuffer.readline()
            if not data[-2::] == SYM_CRLF:
                return NOT_ENOUGH_DATA
            data_len = len(data)
            self.bytes_read += data_len
            marker, chunk = data[0], data[1:-2]
            response: ResponseType = None
            if marker == RESPDataType.SIMPLE_STRING:
                response = chunk
                if decode_bytes and encoding:
                    response = self.try_decode(response, encoding)
            elif marker == RESPDataType.BULK_STRING or marker == RESPDataType.VERBATIM:
                length = int(chunk)
                if length == -1:
                    response = None
                else:
                    if (self.bytes_written - self.bytes_read) < length + 2:
                        self.bytes_read -= data_len
                        return NOT_ENOUGH_DATA
                    data = self.localbuffer.read(length + 2)
                    self.bytes_read += length + 2
                    response = data[:-2]
                    if marker == RESPDataType.VERBATIM:
                        if response[:3] != b"txt":
                            raise InvalidResponse(
                                f"Unexpected verbatim string of type {response[:3]!r}"
                            )
                        response = response[4:]
                    if decode_bytes and encoding:
                        response = self.try_decode(response, encoding)
            elif marker in [RESPDataType.INT, RESPDataType.BIGNUMBER]:
                response = int(chunk)
            elif marker == RESPDataType.DOUBLE:
                response = float(chunk)
            elif marker == RESPDataType.NONE:
                response = None
            elif marker == RESPDataType.BOOLEAN:
                response = chunk[0] == ord(b"t")
            elif (
                marker == RESPDataType.ARRAY
                or marker == RESPDataType.PUSH
                or marker == RESPDataType.MAP
                or marker == RESPDataType.SET
            ):
                length = int(chunk)
                if length >= 0:
                    if marker in {RESPDataType.ARRAY, RESPDataType.PUSH}:
                        self.nodes.append(ListNode(length, marker))
                    elif marker == RESPDataType.MAP:
                        self.nodes.append(DictNode(length))
                    else:
                        self.nodes.append(SetNode(length))
                    if length > 0:
                        continue
            elif marker == RESPDataType.ERROR:
                response = cast(ResponseType, self.parse_error(bytes(chunk).decode()))
            else:
                raise InvalidResponse(
                    f"Protocol Error: {chr(marker)}, {bytes(chunk)!r}"
                )

            if self.nodes:
                if self.nodes[-1].depth > 0:
                    self.nodes[-1].append(response)
                    if self.nodes[-1].depth > 1:
                        continue

                while len(self.nodes) > 1 and self.nodes[-1].depth == 0:
                    self.nodes[-2].append(self.nodes[-1].container)
                    self.nodes.pop()

            if len(self.nodes) == 1 and self.nodes[-1].depth == 0:
                node = self.nodes.pop()
                parsed = UnpackedResponse(node.node_type, node.container)
                break
            if not self.nodes:
                parsed = UnpackedResponse(marker, response)
                break

        if self.bytes_read == self.bytes_written:
            self.localbuffer.seek(0)
            self.localbuffer.truncate()
            self.bytes_read = self.bytes_written = 0
        return parsed

    def parse_error(self, response: str) -> RedisError:
        """
        Parse an error response

        :meta private:
        """
        error_code = response.split(" ")[0]
        if error_code in self.EXCEPTION_CLASSES:
            response = response[len(error_code) + 1 :]
            exception_class = self.EXCEPTION_CLASSES[error_code]

            if isinstance(exception_class, dict):
                options = exception_class.items()
                exception_class = ResponseError
                for err, exc in options:
                    if response.lower().startswith(err):
                        exception_class = exc
                        break
            return exception_class(response)
        return ResponseError(response)



from __future__ import annotations

from typing import Any, cast

from coredis._utils import EncodingInsensitiveDict
from coredis.response._callbacks import (
    ClusterMergeMapping,
    DictCallback,
    ResponseCallback,
)
from coredis.response._utils import flat_pairs_to_dict
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Mapping,
    Optional,
    ResponsePrimitive,
    ResponseType,
    Sequence,
    Tuple,
    Union,
    ValueT,
)


class SampleCallback(
    ResponseCallback[
        List[ValueT],
        List[ValueT],
        Union[Tuple[int, float], Tuple[()]],
    ]
):
    def transform(
        self,
        response: List[ValueT],
        **options: Optional[ValueT],
    ) -> Union[Tuple[int, float], Tuple[()]]:
        return (int(response[0]), float(response[1])) if response else ()


class SamplesCallback(
    ResponseCallback[
        Optional[List[List[ValueT]]],
        Optional[List[List[ValueT]]],
        Union[Tuple[Tuple[int, float], ...], Tuple[()]],
    ],
):
    def transform(
        self,
        response: Optional[List[List[ValueT]]],
        **options: Optional[ValueT],
    ) -> Union[Tuple[Tuple[int, float], ...], Tuple[()]]:
        if response:
            return tuple(
                cast(Tuple[int, float], SampleCallback().transform(r)) for r in response
            )
        return ()


class TimeSeriesInfoCallback(DictCallback[AnyStr, ResponseType]):
    def transform(
        self,
        response: Union[Sequence[ResponseType], Dict[ResponsePrimitive, ResponseType]],
        **options: Optional[ValueT],
    ) -> Dict[AnyStr, ResponseType]:
        dct = EncodingInsensitiveDict(super().transform(response, **options))
        if "labels" in dct:
            dct["labels"] = dict(dct["labels"])
        if "Chunks" in dct:
            dct["Chunks"] = [flat_pairs_to_dict(chunk) for chunk in dct["Chunks"]]
        if "rules" in dct and not isinstance(dct["rules"], dict):
            dct["rules"] = {rule[0]: rule[1:] for rule in dct["rules"]}

        return dict(dct)


class TimeSeriesCallback(
    ResponseCallback[
        ResponseType,
        ResponseType,
        Dict[AnyStr, Tuple[Dict[AnyStr, AnyStr], Tuple[int, float]]],
    ]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Dict[AnyStr, Tuple[Dict[AnyStr, AnyStr], Tuple[int, float]]]:
        if isinstance(response, dict):
            return {k: (v[0], tuple(v[1])) for k, v in response.items()}
        else:
            return {
                r[0]: (dict(r[1]), (r[2][0], float(r[2][1])) if r[2] else tuple())
                for r in response
            }


class TimeSeriesMultiCallback(
    ResponseCallback[
        ResponseType,
        ResponseType,
        Dict[
            AnyStr,
            Tuple[
                Dict[AnyStr, AnyStr], Union[Tuple[Tuple[int, float], ...], Tuple[()]]
            ],
        ],
    ]
):
    def transform(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Dict[
        AnyStr,
        Tuple[Dict[AnyStr, AnyStr], Union[Tuple[Tuple[int, float], ...], Tuple[()]]],
    ]:
        if options.get("grouped"):
            return {
                r[0]: (
                    flat_pairs_to_dict(r[1][0]) if r[1] else {},
                    tuple(SampleCallback().transform(t) for t in r[2]),
                )
                for r in cast(Any, response)
            }
        else:
            return {
                r[0]: (
                    dict(r[1]),
                    tuple(SampleCallback().transform(t) for t in r[2]),
                )
                for r in cast(Any, response)
            }

    def transform_3(
        self, response: ResponseType, **options: Optional[ValueT]
    ) -> Dict[
        AnyStr,
        Tuple[Dict[AnyStr, AnyStr], Union[Tuple[Tuple[int, float], ...], Tuple[()]]],
    ]:
        if isinstance(response, dict):
            if options.get("grouped"):
                return {
                    k: (
                        r[0],
                        tuple(SampleCallback().transform(t) for t in r[-1]),
                    )
                    for k, r in response.items()
                }
            else:
                return {
                    k: (
                        r[0],
                        tuple(SampleCallback().transform(t) for t in r[-1]),
                    )
                    for k, r in response.items()
                }
        else:
            return self.transform(response, **options)


class ClusterMergeTimeSeries(ClusterMergeMapping[AnyStr, Tuple[Any, ...]]):
    def __init__(self) -> None:
        self.value_combine = self.merge

    def combine(
        self,
        responses: Mapping[str, Dict[AnyStr, Tuple[Any, ...]]],
        **kwargs: Optional[ValueT],
    ) -> Dict[AnyStr, Tuple[Any, ...]]:
        if not kwargs.get("grouped"):
            return super().combine(responses, **kwargs)
        raise NotImplementedError(
            "Unable to merge response from multiple cluster nodes when used with grouping"
        )

    def merge(
        self, values: Any
    ) -> Tuple[Dict[AnyStr, AnyStr], Tuple[Tuple[int, float], ...]]:
        merged_labels: Dict[AnyStr, AnyStr] = {}
        merged_series: Tuple[Tuple[int, float], ...] = ()
        for value in values:
            merged_labels.update(value[0])
            merged_series = merged_series + value[1]
        return merged_labels, tuple(merged_series)



from __future__ import annotations

import dataclasses
from typing import Any, NamedTuple

from coredis._json import json
from coredis.typing import (
    AnyStr,
    Dict,
    Generic,
    List,
    Optional,
    ResponsePrimitive,
    ResponseType,
    Set,
    StringT,
    Tuple,
    Union,
)


@dataclasses.dataclass
class SearchDocument(Generic[AnyStr]):
    """
    Search document as returned by `FT.SEARCH <https://redis.io/commands/ft.search>`__
    """

    #: Document id
    id: StringT
    #: Search score if the :paramref:`~coredis.modules.search.Search.search.withscores`
    #: option was used
    score: Optional[float]
    #: Explanation of the score if the
    #:  :paramref:`~coredis.modules.search.Search.search.explainscore` option was used
    score_explanation: Optional[List[AnyStr]]
    #: Payload associated with the document if
    #:  :paramref:`~coredis.modules.search.Search.search.withpayloads` was used
    payload: Optional[StringT]
    sortkeys: Optional[StringT]
    #: Mapping of properties returned for the document
    properties: Dict[AnyStr, ResponseType]


@dataclasses.dataclass
class SearchResult(Generic[AnyStr]):
    """
    Search results as returned by `FT.SEARCH <https://redis.io/commands/ft.search>`__
    """

    #: The total number of results found for the query
    total: int
    #: The documents returned by the query
    documents: Tuple[SearchDocument[AnyStr], ...]


@dataclasses.dataclass
class SearchAggregationResult(Generic[AnyStr]):
    """
    Search aggregations as returned by `FT.AGGREGATE <https://redis.io/commands/ft.aggregate>`__
    """

    #: The aggregation results
    results: List[Dict[StringT, ResponseType]]
    #: The cursor id if :paramref:`~coredis.modules.search.aggregate.with_cursor` was `True`
    cursor: Optional[int]

    def __post_init__(self) -> None:
        for idx, result in enumerate(self.results):
            json_key = b"$" if b"$" in result else "$" if "$" in result else None
            if json_key:
                self.results[idx] = json.loads(result.pop(json_key))


@dataclasses.dataclass
class AutocompleteSuggestion(Generic[AnyStr]):
    """
    Autocomplete suggestion as returned by `FT.SUGGET <https://redis.io/commands/ft.sugget>`__
    """

    #: the suggestion string
    string: AnyStr
    #: the score of the suggestion if
    #:  :paramref:`~coredis.modules.autocomplete.Autocomplete.sugget.withscores` was used
    score: Optional[float]
    #: the payload associated with the suggestion if
    #:  :paramref:`~coredis.modules.autocomplete.Autocomplete.sugget.withpayloads` was used
    payload: Optional[AnyStr]


#: Type alias for valid python types that can be represented as json
JsonType = Union[str, int, float, bool, None, Dict[str, Any], List[Any]]


@dataclasses.dataclass
class GraphNode(Generic[AnyStr]):
    """
    Representation of a graph node
    """

    #: The node's internal ID
    id: int
    #: A set of labels associated with the node
    labels: Set[AnyStr]
    #: Mapping of property names to values
    properties: Dict[AnyStr, ResponseType]


@dataclasses.dataclass
class GraphRelation(Generic[AnyStr]):
    """
    Representation of a relation between two nodes
    """

    #: The relationship's internal ID
    id: int
    #: Relation type
    type: AnyStr
    #: Source node ID
    src_node: int
    #: Destination node ID
    destination_node: int
    #: Mapping of all properties the relation possesses
    properties: Dict[AnyStr, ResponseType]


@dataclasses.dataclass
class GraphPath(Generic[AnyStr]):
    """
    Representation of a graph path
    """

    #: The nodes in the path
    nodes: List[GraphNode[AnyStr]]
    #: The relations in the path
    relations: List[GraphRelation[AnyStr]]

    NULL_NODE = GraphNode[AnyStr](0, set(), {})

    @property
    def path(self) -> Tuple[Union[GraphNode[AnyStr], GraphRelation[AnyStr]], ...]:
        """
        The path as a tuple of nodes and relations
        """
        if self.nodes and self.relations:
            return tuple(
                [
                    item
                    for pair in zip(self.nodes, self.relations + [self.NULL_NODE])
                    for item in pair
                ][:-1]
            )
        return ()


@dataclasses.dataclass
class GraphQueryResult(Generic[AnyStr]):
    """
    Response from `GRAPH.QUERY <https://redis.io/commands/graph.query>`__
    """

    #: List of entries in the response header
    header: Tuple[AnyStr, ...]
    #: The result set from the query
    result_set: Tuple[
        Union[
            ResponsePrimitive,
            List[
                Union[
                    ResponsePrimitive,
                    GraphNode[AnyStr],
                    GraphRelation[AnyStr],
                    GraphPath[AnyStr],
                ]
            ],
        ],
        ...,
    ]
    #: Mapping of query statistics
    stats: Dict[str, ResponsePrimitive]


class GraphSlowLogInfo(NamedTuple):
    """
    Response from `GRAPH.SLOWLOG <https://redis.io/commands/graph.slowlog>`__
    """

    #: The unix timestamp at which the logged command was processed.
    start_time: int
    #: The array composing the arguments of the command.
    command: StringT
    #: query name
    query: StringT
    #: The amount of time needed for its execution, in microseconds.
    duration: float



from __future__ import annotations

from coredis._utils import EncodingInsensitiveDict, nativestr
from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict
from coredis.response.types import Command
from coredis.typing import (
    AnyStr,
    Dict,
    List,
    Optional,
    ResponsePrimitive,
    ResponseType,
    Set,
    ValueT,
)


class CommandCallback(
    ResponseCallback[List[ResponseType], List[ResponseType], Dict[str, Command]]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Dict[str, Command]:
        commands: Dict[str, Command] = {}

        for command in response:
            assert isinstance(command, list)

            if command:
                name = nativestr(command[0])

                if len(command) >= 6:
                    commands[name] = {
                        "name": command[0],
                        "arity": command[1],
                        "flags": command[2],
                        "first-key": command[3],
                        "last-key": command[4],
                        "step": command[5],
                        "acl-categories": None,
                        "tips": None,
                        "key-specifications": None,
                        "sub-commands": None,
                    }

                if len(command) >= 7:
                    commands[name]["acl-categories"] = command[6]

                if len(command) >= 8:
                    commands[name]["tips"] = command[7]
                    commands[name]["key-specifications"] = command[8]
                    commands[name]["sub-commands"] = command[9]

        return commands


class CommandKeyFlagCallback(
    ResponseCallback[List[ResponseType], List[ResponseType], Dict[AnyStr, Set[AnyStr]]]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Dict[AnyStr, Set[AnyStr]]:
        return {k[0]: set(k[1]) for k in response}


class CommandDocCallback(
    ResponseCallback[
        List[ResponseType],
        Dict[ResponsePrimitive, ResponseType],
        Dict[AnyStr, Dict[AnyStr, ResponseType]],
    ]
):
    def transform(
        self, response: List[ResponseType], **options: Optional[ValueT]
    ) -> Dict[AnyStr, Dict[AnyStr, ResponseType]]:
        cmd_mapping = flat_pairs_to_dict(response)
        for cmd, doc in cmd_mapping.items():
            cmd_mapping[cmd] = EncodingInsensitiveDict(flat_pairs_to_dict(doc))
            cmd_mapping[cmd]["arguments"] = [
                flat_pairs_to_dict(arg) for arg in cmd_mapping[cmd].get("arguments", [])
            ]
        return cmd_mapping

    def transform_3(
        self,
        response: Dict[ResponsePrimitive, ResponseType],
        **options: Optional[ValueT],
    ) -> Dict[AnyStr, Dict[AnyStr, ResponseType]]:
        return response  # noqa
